package com.sun.org.apache.bcel.internal.util;

import java.io.IOException;
import com.sun.org.apache.bcel.internal.classfile.ClassParser;
import com.sun.org.apache.bcel.internal.Repository;
import com.sun.org.apache.bcel.internal.generic.ArrayType;
import com.sun.org.apache.bcel.internal.generic.Type;
import java.util.Locale;
import com.sun.org.apache.bcel.internal.Const;
import com.sun.org.apache.bcel.internal.generic.MethodGen;
import com.sun.org.apache.bcel.internal.classfile.ConstantValue;
import com.sun.org.apache.bcel.internal.classfile.Method;
import com.sun.org.apache.bcel.internal.classfile.Field;
import com.sun.org.apache.bcel.internal.classfile.Utility;
import java.io.OutputStream;
import com.sun.org.apache.bcel.internal.generic.ConstantPoolGen;
import java.io.PrintWriter;
import com.sun.org.apache.bcel.internal.classfile.JavaClass;
import com.sun.org.apache.bcel.internal.classfile.EmptyVisitor;

public class BCELifier extends EmptyVisitor
{
    private static final String BASE_PACKAGE;
    private static final String CONSTANT_PREFIX;
    private final JavaClass _clazz;
    private final PrintWriter _out;
    private final ConstantPoolGen _cp;
    
    public BCELifier(JavaClass clazz, OutputStream out) {
        this._clazz = clazz;
        this._out = new PrintWriter(out);
        this._cp = new ConstantPoolGen(this._clazz.getConstantPool());
    }
    
    public void start() {
        this.visitJavaClass(this._clazz);
        this._out.flush();
    }
    
    @Override
    public void visitJavaClass(JavaClass clazz) {
        String class_name = clazz.getClassName();
        String super_name = clazz.getSuperclassName();
        String package_name = clazz.getPackageName();
        String inter = Utility.printArray(clazz.getInterfaceNames(), false, true);
        if (!"".equals(package_name)) {
            class_name = class_name.substring(package_name.length() + 1);
            this._out.println("package " + package_name + ";");
            this._out.println();
        }
        this._out.println("import " + BCELifier.BASE_PACKAGE + ".generic.*;");
        this._out.println("import " + BCELifier.BASE_PACKAGE + ".classfile.*;");
        this._out.println("import " + BCELifier.BASE_PACKAGE + ".*;");
        this._out.println("import java.io.*;");
        this._out.println();
        this._out.println("public class " + class_name + "Creator {");
        this._out.println("  private InstructionFactory _factory;");
        this._out.println("  private ConstantPoolGen    _cp;");
        this._out.println("  private ClassGen           _cg;");
        this._out.println();
        this._out.println("  public " + class_name + "Creator() {");
        this._out.println("    _cg = new ClassGen(\"" + ("".equals(package_name) ? class_name : (package_name + "." + class_name)) + "\", \"" + super_name + "\", \"" + clazz.getSourceFileName() + "\", " + printFlags(clazz.getAccessFlags(), BCELifier.FLAGS.CLASS) + ", new String[] { " + inter + " });");
        this._out.println();
        this._out.println("    _cp = _cg.getConstantPool();");
        this._out.println("    _factory = new InstructionFactory(_cg, _cp);");
        this._out.println("  }");
        this._out.println();
        this.printCreate();
        Field[] fields = clazz.getFields();
        if (fields.length > 0) {
            this._out.println("  private void createFields() {");
            this._out.println("    FieldGen field;");
            for (Field field : fields) {
                field.accept(this);
            }
            this._out.println("  }");
            this._out.println();
        }
        Method[] methods = clazz.getMethods();
        for (int i = 0; i < methods.length; ++i) {
            this._out.println("  private void createMethod_" + i + "() {");
            methods[i].accept(this);
            this._out.println("  }");
            this._out.println();
        }
        this.printMain();
        this._out.println("}");
    }
    
    private void printCreate() {
        this._out.println("  public void create(OutputStream out) throws IOException {");
        Field[] fields = this._clazz.getFields();
        if (fields.length > 0) {
            this._out.println("    createFields();");
        }
        Method[] methods = this._clazz.getMethods();
        for (int i = 0; i < methods.length; ++i) {
            this._out.println("    createMethod_" + i + "();");
        }
        this._out.println("    _cg.getJavaClass().dump(out);");
        this._out.println("  }");
        this._out.println();
    }
    
    private void printMain() {
        String class_name = this._clazz.getClassName();
        this._out.println("  public static void main(String[] args) throws Exception {");
        this._out.println("    " + class_name + "Creator creator = new " + class_name + "Creator();");
        this._out.println("    creator.create(new FileOutputStream(\"" + class_name + ".class\"));");
        this._out.println("  }");
    }
    
    @Override
    public void visitField(Field field) {
        this._out.println();
        this._out.println("    field = new FieldGen(" + printFlags(field.getAccessFlags()) + ", " + printType(field.getSignature()) + ", \"" + field.getName() + "\", _cp);");
        ConstantValue cv = field.getConstantValue();
        if (cv != null) {
            String value = cv.toString();
            this._out.println("    field.setInitValue(" + value + ")");
        }
        this._out.println("    _cg.addField(field.getField());");
    }
    
    @Override
    public void visitMethod(Method method) {
        MethodGen mg = new MethodGen(method, this._clazz.getClassName(), this._cp);
        this._out.println("    InstructionList il = new InstructionList();");
        this._out.println("    MethodGen method = new MethodGen(" + printFlags(method.getAccessFlags(), BCELifier.FLAGS.METHOD) + ", " + printType(mg.getReturnType()) + ", " + printArgumentTypes(mg.getArgumentTypes()) + ", new String[] { " + Utility.printArray(mg.getArgumentNames(), false, true) + " }, \"" + method.getName() + "\", \"" + this._clazz.getClassName() + "\", il, _cp);");
        this._out.println();
        BCELFactory factory = new BCELFactory(mg, this._out);
        factory.start();
        this._out.println("    method.setMaxStack();");
        this._out.println("    method.setMaxLocals();");
        this._out.println("    _cg.addMethod(method.getMethod());");
        this._out.println("    il.dispose();");
    }
    
    static String printFlags(int flags) {
        return printFlags(flags, BCELifier.FLAGS.UNKNOWN);
    }
    
    public static String printFlags(int flags, BCELifier.FLAGS location) {
        if (flags == 0) {
            return "0";
        }
        StringBuilder buf = new StringBuilder();
        for (int i = 0, pow = 1; pow <= 16384; pow <<= 1, ++i) {
            if ((flags & pow) != 0x0) {
                if (pow == 32 && location == BCELifier.FLAGS.CLASS) {
                    buf.append(BCELifier.CONSTANT_PREFIX + "ACC_SUPER | ");
                }
                else if (pow == 64 && location == BCELifier.FLAGS.METHOD) {
                    buf.append(BCELifier.CONSTANT_PREFIX + "ACC_BRIDGE | ");
                }
                else if (pow == 128 && location == BCELifier.FLAGS.METHOD) {
                    buf.append(BCELifier.CONSTANT_PREFIX + "ACC_VARARGS | ");
                }
                else if (i < Const.ACCESS_NAMES_LENGTH) {
                    buf.append(BCELifier.CONSTANT_PREFIX + "ACC_").append(Const.getAccessName(i).toUpperCase(Locale.ENGLISH)).append(" | ");
                }
                else {
                    buf.append(String.format(BCELifier.CONSTANT_PREFIX + "ACC_BIT %x | ", new Object[] { Integer.valueOf(pow) }));
                }
            }
        }
        String str = buf.toString();
        return str.substring(0, str.length() - 3);
    }
    
    static String printArgumentTypes(Type[] arg_types) {
        if (arg_types.length == 0) {
            return "Type.NO_ARGS";
        }
        StringBuilder args = new StringBuilder();
        for (int i = 0; i < arg_types.length; ++i) {
            args.append(printType(arg_types[i]));
            if (i < arg_types.length - 1) {
                args.append(", ");
            }
        }
        return "new Type[] { " + args.toString() + " }";
    }
    
    static String printType(Type type) {
        return printType(type.getSignature());
    }
    
    static String printType(String signature) {
        Type type = Type.getType(signature);
        byte t = type.getType();
        if (t <= 12) {
            return "Type." + Const.getTypeName((int)t).toUpperCase(Locale.ENGLISH);
        }
        if (type.toString().equals("java.lang.String")) {
            return "Type.STRING";
        }
        if (type.toString().equals("java.lang.Object")) {
            return "Type.OBJECT";
        }
        if (type.toString().equals("java.lang.StringBuffer")) {
            return "Type.STRINGBUFFER";
        }
        if ((type instanceof ArrayType)) {
            ArrayType at = (ArrayType)type;
            return "new ArrayType(" + printType(at.getBasicType()) + ", " + at.getDimensions() + ")";
        }
        return "new ObjectType(\"" + Utility.signatureToString(signature, false) + "\")";
    }
    
    public static void main(String[] argv) throws Exception {
        if (argv.length != 1) {
            System.out.println("Usage: BCELifier classname");
            System.out.println("\tThe class must exist on the classpath");
            return;
        }
        JavaClass java_class = getJavaClass(argv[0]);
        BCELifier bcelifier = new BCELifier(java_class, System.out);
        bcelifier.start();
        return;
    }
    
    static JavaClass getJavaClass(String name) throws ClassNotFoundException, IOException {
        JavaClass java_class;
        if ((java_class = Repository.lookupClass(name)) == null) {
            java_class = new ClassParser(name).parse();
            return java_class;
        }
        return java_class;
    }
    
    static {
        BASE_PACKAGE = Const.class.getPackage().getName();
        CONSTANT_PREFIX = Const.class.getSimpleName() + ".";
    }
    
    public enum FLAGS
    {
        UNKNOWN, 
        CLASS, 
        METHOD;
    }
}
